@@ -2,7 +2,7 @@
 /* http-neon-method.c - The HTTP method implementation for the GNOME Virtual 
    File System using the neon http/webdav library.
 
-   Copyright (C) 2004 Christian Kellner <gicmo@gnome-de.org>
+   Copyright (C) 2004 Christian Kellner <gicmo@gnome.org>
    
    The Gnome Library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public License as
@@ -71,6 +71,7 @@
 #define DEFAULT_HTTP_PORT 	80
 #define DEFAULT_HTTPS_PORT 	443
 
+#define HTTP_DIR_MIME_TYPE      "x-directory/webdav"
 
 /* ************************************************************************** */
 /* module interfaces */
@@ -82,25 +83,25 @@ void               vfs_module_shutdown  (GnomeVFSMethod *method);
 /* ************************************************************************** */
 /* DEBUGING stuff */
 
+/* #define DEBUG_HTTP_ENABLE 1 */
+#undef DEBUG_HTTP_ENABLE
 #ifdef DEBUG_HTTP_ENABLE
 
-void http_debug_printf(char *func, char *fmt, ...) G_GNUC_PRINTF (2,3);
+void http_debug_printf(const char *func, const char *fmt, ...) G_GNUC_PRINTF (2,3);
 
 
-/* #define DEBUG_HTTP_3(fmt, ...) http_debug_printf (__PRETTY_FUNCTION__, fmt, ##__VA_ARGS__) */
-/* #define DEBUG_HTTP_2(fmt, ...) http_debug_printf (__PRETTY_FUNCTION__, fmt, ##__VA_ARGS__) */
+#define DEBUG_HTTP_3(fmt, ...) http_debug_printf (__PRETTY_FUNCTION__, fmt, ##__VA_ARGS__) 
+#define DEBUG_HTTP_2(fmt, ...) http_debug_printf (__PRETTY_FUNCTION__, fmt, ##__VA_ARGS__) 
 #define DEBUG_HTTP(fmt, ...) http_debug_printf (__PRETTY_FUNCTION__, fmt, ##__VA_ARGS__)
 #define DEBUG_HTTP_FUNC(_enter) http_debug_printf (_enter ? "+++" : "---", "%s",__PRETTY_FUNCTION__)
-#define DEBUG_HTTP_3(fmt, ...)	
-#define DEBUG_HTTP_2(fmt, ...)
 
 
 #define DEBUG_HTTP_CONTEXT(c) http_debug_printf (__PRETTY_FUNCTION__,       \
-						  "[Context] Session :%p, path: %s", \
-						  c->session, c->path)
+						  "[Context] Session :%p, path: %s, dav_mode %d", \
+						  c->session, c->path, c->dav_mode)
 
 void
-http_debug_printf (char *func, char *fmt, ...)
+http_debug_printf (const char *func, const char *fmt, ...)
 {
 	va_list args;
 	gchar * out;
@@ -274,22 +275,31 @@ resolve_result (int result, ne_request *request)
 /* ************************************************************************** */
 /* Utility functions */
 
+typedef enum {
+	
+	DAV_CLASS_NOT_SET = -1,
+	NO_DAV            =  0,
+	DAV_CLASS_1       =  1,
+	DAV_CLASS_2       =  2
+
+} DavClass;
+
 typedef struct {
 	
 	char      *scheme;
 	gboolean   ssl;
 	char      *alias;
+	gboolean   is_dav;
 	   
 } MethodSchemes;
 
+static MethodSchemes supported_schemes[] = {
 
-MethodSchemes supported_schemes[] = {
-
-	{"http",  FALSE, "http" },
-	{"dav",   FALSE, "http" },
-	{"https", TRUE,  "https"},
-	{"davs",  TRUE,  "https"},
-	{NULL,    FALSE         }	   
+	{"http",  FALSE, "http",  FALSE},
+	{"dav",   FALSE, "http",  TRUE},
+	{"https", TRUE,  "https", FALSE},
+	{"davs",  TRUE,  "https", TRUE},
+	{NULL,    FALSE,          NULL}	   
 };
 
 static char *
@@ -311,6 +321,22 @@ resolve_alias (const char *scheme)
 	return iter->alias;
 }
 
+static gboolean
+scheme_is_dav (GnomeVFSURI *uri)
+{
+	const char *scheme;
+
+	scheme = gnome_vfs_uri_get_scheme (uri);
+	
+	if (scheme == NULL)
+		return FALSE;
+
+	if (! g_ascii_strcasecmp (scheme, "dav") ||
+	    ! g_ascii_strcasecmp (scheme, "davs"))
+		return TRUE;
+
+	return FALSE;
+}
 
 static guint
 http_session_uri_hash (gconstpointer v)
@@ -963,10 +989,10 @@ static void
 set_dav_class (void *userdata, const char *value)
 {
 	char *tokens = ne_strdup(value), *pnt = tokens;
-	gint *dav_class = userdata;
+        DavClass *dav_class = userdata;
     
 	DEBUG_HTTP_3 ("parsing dav: %s", tokens); 
-	
+
 	do {
 		char *tok = ne_qtoken (&pnt, ',',  "\"'");
 		if (!tok) break;
@@ -1087,7 +1113,8 @@ set_content_range (void *ud, const char *value)
     	}
 }
 
-
+/* function for setting the etag value
+   we dont use it at the moment
 static void 
 set_etag (void **etag, const char *value)
 {		
@@ -1096,6 +1123,30 @@ set_etag (void **etag, const char *value)
 	else
 		*etag = NULL;
 }
+*/
+
+static void
+add_default_header_handlers (ne_request *req, GnomeVFSFileInfo *info)
+{
+
+	ne_add_response_header_handler (req, "Last-Modified", 
+					(ne_header_handler) set_last_modified,
+		   			info);
+	
+	ne_add_response_header_handler (req, "Content-Length", 
+					(ne_header_handler) set_content_length,
+					info);
+	
+	ne_add_response_header_handler (req, "Content-Type",
+					(ne_header_handler) set_content_type,
+					info);
+	
+	ne_add_response_header_handler (req, "Date",
+					(ne_header_handler) set_access_time,
+					info);
+       
+}
+
 
 /* ************************************************************************** */
 /* Propfind request handlers */
@@ -1105,7 +1156,7 @@ static const ne_propname file_info_props[] = {
 	{"DAV:", "resourcetype"},
 	{"DAV:", "getcontenttype"},
 	{"DAV:", "getcontentlength"},
-	{"DAV:", "getetag"},
+	/*      {"DAV:", "getetag"}, */
 	/*	{"http://apache.org/dav/props/", "executable"}, */
 	{NULL}
 };
@@ -1137,15 +1188,23 @@ propfind_context_init (PropfindContext *pfctx)
 static void
 propfind_context_clear (PropfindContext *pfctx)
 {
-	if (pfctx->target != NULL)
-		gnome_vfs_file_info_unref (pfctx->target);
+	GList *l;
 	
+	if (pfctx->target != NULL) {
+		gnome_vfs_file_info_unref (pfctx->target);
+		pfctx->target = NULL;
+	}
+		
 	if (pfctx->children) {
-		g_list_free (pfctx->children);
+		l = gnome_vfs_file_info_list_unref (pfctx->children);
+		g_list_free (l);
+		pfctx->children = NULL;
 	}
 	
-	if (pfctx->etag)
+	if (pfctx->etag) {
 		g_free (pfctx->etag);
+		pfctx->etag = NULL;
+	}
 }
 
 
@@ -1177,16 +1236,20 @@ propfind_result (void *userdata, const char *href, const ne_prop_result_set *set
 	info->name = g_path_get_basename (unesc_path);
 	NE_FREE (unesc_path);
 	
+	DEBUG_HTTP_2 ("Comparing: \n\t[%s] \n\t[%s]", ctx->path, uri.path);
 	if (ne_path_compare (ctx->path, uri.path) == 0) {
 		DEBUG_HTTP_3 ("target");
 		ctx->target = info;
 		
 		/* Set the etag on target */
+		/* We are not requesting the etag information at 
+		 * the moment so don't even check for it 
 		value = ne_propset_value (set, &file_info_props[5]);
 		
 		if (value != NULL)
 			ctx->etag = g_strdup (value);
-		
+	
+		*/	
 	} else {
 		ctx->children = g_list_append (ctx->children, info);
 	}
@@ -1214,7 +1277,7 @@ propfind_result (void *userdata, const char *href, const ne_prop_result_set *set
 	info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
 	
 	if (value != NULL && strstr (value, "collection")) {
-		info->mime_type = g_strdup ("x-directory/webdav");
+		info->mime_type = g_strdup (HTTP_DIR_MIME_TYPE);
 		info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
 		info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
 		
@@ -1253,7 +1316,7 @@ propfind_result (void *userdata, const char *href, const ne_prop_result_set *set
 
 G_LOCK_DEFINE (nst_lock);
 static GHashTable *neon_session_table;
-guint nst_tid;
+static guint nst_tid;
 
 typedef struct {
 	
@@ -1280,6 +1343,7 @@ neon_session_pool_destroy (NeonSessionPool *pool)
 	g_list_free (pool->unused_sessions);
 	
 	gnome_vfs_uri_unref (pool->uri);
+	g_free (pool);
 }
 
 
@@ -1384,10 +1448,12 @@ neon_session_pool_lookup (GnomeVFSURI *uri)
 	DEBUG_HTTP ("[Session Pool] Searching (%d)", 
 		    g_hash_table_size (neon_session_table));
 	
+	session = NULL;
 	pool = g_hash_table_lookup (neon_session_table, uri);
 	
 	/* search in session pool */
 	if (pool != NULL && pool->unused_sessions) {
+
 		session = pool->unused_sessions->data;
 		pool->unused_sessions = g_list_remove (pool->unused_sessions,
 						       session);
@@ -1395,59 +1461,112 @@ neon_session_pool_lookup (GnomeVFSURI *uri)
 			    g_list_length (pool->unused_sessions));
 		
 		g_get_current_time (&(pool->last_used));
-
-		G_UNLOCK (nst_lock);
-		return session;
 	}
 	
 	G_UNLOCK (nst_lock);
-	return NULL;
+	return session;
 }
 
 /* ************************************************************************** */
 /* Additional Headers */
-#if 0
-/* Does not work because the neon library doesnt provide as all the headers
-   working around that is a bigger thing */
-
-typedef struct {
-	
+static int 
+neon_return_headers (ne_request *req, void *userdata, const ne_status *status)
+{	
+	GnomeVFSModuleCallbackReceivedHeadersIn in_args;
+	GnomeVFSModuleCallbackReceivedHeadersOut out_args;
+	GList **headers, *iter;
 	GnomeVFSURI *uri;
-	gboolean sent;
+	ne_session *session;
+
+	DEBUG_HTTP_FUNC (1);
 	
-} AddHeadersContext;
+	session = ne_get_session (req);
+	
+	if (ne_get_request_private (req, "Headers Returned"))
+		return 0;
+	
+	headers = ne_get_request_private (req, "Headers");
+	uri = ne_get_session_private (session, "GnomeVFSURI");
+
+	memset (&in_args, 0, sizeof (in_args));
+	memset (&out_args, 0, sizeof (out_args));
+	
+	in_args.uri = uri;
+	in_args.headers = *headers;
+
+	gnome_vfs_module_callback_invoke (GNOME_VFS_MODULE_CALLBACK_HTTP_RECEIVED_HEADERS,
+					&in_args, sizeof (in_args),
+					&out_args, sizeof (out_args));
+	
+	for (iter = *headers; iter; iter = iter->next) {
+		DEBUG_HTTP_3 ("Headers returned %s,", (char *) iter->data);
+		g_free (iter->data);
+	}
+
+	g_list_free (*headers);
+	g_free (headers);
+
+	ne_set_request_private (req, "Headers Returned", "TRUE");
+	
+	DEBUG_HTTP_FUNC (0);
+	
+	return 0;
+}
+
+static void 
+neon_header_catcher (void *userdata, const char *value)
+{
+	GList **headers = (GList **) userdata;
+
+	DEBUG_HTTP_3 ("Catching Header %s,", (char *) value);
+
+	*headers = g_list_prepend (*headers, g_strdup (value));	
+}
 
 
 static void 
-neon_handle_additional_headers (ne_request *req, 
-				void *userdata, 
-				const char *method, 
-				const char *requri)
+neon_setup_headers (ne_request *req, void *userdata, ne_buffer *header)
 {
-	AddHeadersContext *adhctx = (AddHeadersContext *) userdata;
+	GnomeVFSModuleCallbackAdditionalHeadersIn in_args;
+	GnomeVFSModuleCallbackAdditionalHeadersOut out_args;
+	GList *iter, **headers;
+	gboolean ret;
+	GnomeVFSURI *uri;
+	ne_session *session;
+
+	DEBUG_HTTP_FUNC (1);
 	
-	if (! invoke_callback_send_additional_headers (uri, &list))
-		return;
+	headers = g_new0 (GList *, 1);
+
+	ne_set_request_private (req, "Headers", headers);
+	ne_add_response_header_catcher (req, neon_header_catcher, headers);
+
+	session = ne_get_session (req);
+	uri = ne_get_session_private (session, "GnomeVFSURI");
+
+	memset (&in_args, 0, sizeof (in_args));
+	memset (&out_args, 0, sizeof (out_args));
+
+	in_args.uri = uri;
+
+	ret = gnome_vfs_module_callback_invoke (GNOME_VFS_MODULE_CALLBACK_HTTP_SEND_ADDITIONAL_HEADERS,
+						&in_args, sizeof (in_args),
+						&out_args, sizeof (out_args));
 	
-	GList *iter;
-	for (iter = list; iter; iter = iter->next) {
-		char *header = iter->data;
-		char *colon  = g_strrstr (header, ":");
-		AcHeaderContext acctx = g_new0 (AcHeaderContext, 1);
+	for (iter = out_args.headers; iter; iter = iter->next) {
 		
-		if (colon != NULL) {
-			colon = '\0';
-			ne_add_request_header (req, header, colon + 1);
-		}
+		if (ret) {
+			ne_buffer_zappend (header, iter->data);
+			DEBUG_HTTP_3 ("Adding header %s,", (char *) iter->data);
+		}		
 		
-		g_free (header);
-		iter->data = NULL;
+		g_free (iter->data);
 	}
 	
-	g_list_free (list);
-
+	g_list_free (out_args.headers);
+	
+	DEBUG_HTTP_FUNC (0);
 }
-#endif
 
 /* ************************************************************************** */
 /* Http context */
@@ -1459,11 +1578,12 @@ typedef struct {
 	const char *scheme;
 	gboolean ssl;
 	
-	guint dav_class;
+	DavClass dav_class;
 	HttpMethods methods;
 	
 	ne_session *session;
-	
+
+	gboolean dav_mode;
 	gboolean redirected;
 	guint    redir_count;
 	
@@ -1487,6 +1607,9 @@ http_aquire_connection (HttpContext *context)
 	session = neon_session_pool_lookup (context->uri);
 	
 	if (session != NULL) {
+			
+		ne_set_session_private (session, "GnomeVFSURI", context->uri);
+	
 		context->session = session;
 		return GNOME_VFS_OK;
 	}
@@ -1520,8 +1643,12 @@ http_aquire_connection (HttpContext *context)
 				 http_auth_info_free, basic_auth);
 
 	ne_redirect_register (session);
-	
-		
+
+	/* Headers stuff */
+	ne_set_session_private (session, "GnomeVFSURI", context->uri);
+	ne_hook_pre_send (session, neon_setup_headers, NULL);
+	ne_hook_post_send (session, neon_return_headers, NULL);
+
 	if (proxy_for_uri (top_uri, &proxy)) {
 		HttpAuthInfo *proxy_auth;
 		
@@ -1558,6 +1685,8 @@ http_context_free (HttpContext *context)
 		neon_session_pool_insert (context->uri, context->session);
 		context->session = NULL;
 	}
+
+	g_free (context->path);
 	
 	gnome_vfs_uri_unref (context->uri);
 	
@@ -1573,20 +1702,20 @@ http_context_set_uri (HttpContext *context, GnomeVFSURI *uri)
 	if (context->uri)
 		gnome_vfs_uri_unref (context->uri);
 	
-	if (context->path == NULL)
+	if (context->path != NULL)
 		g_free (context->path);
 	
-	context->uri = gnome_vfs_uri_ref (uri);
+	context->uri = gnome_vfs_uri_dup (uri);
 	context->scheme = resolve_alias (gnome_vfs_uri_get_scheme (uri));
 	
-	if (gnome_vfs_uri_get_host_port (uri) == 0) {
+	if (gnome_vfs_uri_get_host_port (context->uri) == 0) {
 		if (g_str_equal (context->scheme, "https"))
-			gnome_vfs_uri_set_host_port (uri, DEFAULT_HTTPS_PORT);
+			gnome_vfs_uri_set_host_port (context->uri, DEFAULT_HTTPS_PORT);
 		else 
-			gnome_vfs_uri_set_host_port (uri, DEFAULT_HTTP_PORT);
+			gnome_vfs_uri_set_host_port (context->uri, DEFAULT_HTTP_PORT);
 	}
 	
-	uri_string = gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_USER_NAME
+	uri_string = gnome_vfs_uri_to_string (context->uri, GNOME_VFS_URI_HIDE_USER_NAME
 					      | GNOME_VFS_URI_HIDE_PASSWORD
 					      | GNOME_VFS_URI_HIDE_HOST_NAME
 					      | GNOME_VFS_URI_HIDE_HOST_PORT
@@ -1596,7 +1725,9 @@ http_context_set_uri (HttpContext *context, GnomeVFSURI *uri)
 		uri_string = g_strdup ("/");
 	}
 	
-	context->path = uri_string;
+	context->path      = uri_string;
+	context->methods   = 0;
+	context->dav_class = DAV_CLASS_NOT_SET;
 	
 }
 
@@ -1613,7 +1744,7 @@ http_context_open (GnomeVFSURI *uri, HttpContext **context)
 	ctx = g_new0 (HttpContext, 1);
 	
 	http_context_set_uri (ctx, uri);
-
+	
 	if (ctx->scheme == NULL) {
 		http_context_free (ctx);
 		return GNOME_VFS_ERROR_NOT_SUPPORTED;
@@ -1627,8 +1758,10 @@ http_context_open (GnomeVFSURI *uri, HttpContext **context)
 		return result;
 	}
 
-	*context = ctx;
+	ctx->dav_mode = scheme_is_dav (uri);
 	
+	*context = ctx;
+
 	return GNOME_VFS_OK;
 }
 
@@ -1666,7 +1799,7 @@ http_follow_redirect (HttpContext *context)
 	} else {
 		
 		http_context_set_uri (context, new_uri);
-		
+		ne_set_session_private (context->session, "GnomeVFSURI", context->uri);
 		result = GNOME_VFS_OK;
 	}
 	DEBUG_HTTP ("[Redirec] Redirect result: %s", 
@@ -1678,22 +1811,20 @@ http_follow_redirect (HttpContext *context)
 /* Http operations */
 
 static GnomeVFSResult
-http_get_file_info (HttpContext *context, GnomeVFSFileInfo *info, char **etag)
+http_get_file_info (HttpContext *context, GnomeVFSFileInfo *info)
 {
 	GnomeVFSResult result;
 	PropfindContext pfctx;
 	ne_propfind_handler *pfh;
-	char *_etag;
 	ne_request *req;
 	int res;
 
 	DEBUG_HTTP_CONTEXT (context);
 	
 	/* no dav server */
-	if (context->dav_class == -1) 
+	if (context->dav_mode == FALSE || context->dav_class == NO_DAV) 
 		goto head_start;
  	
-	_etag = NULL;
 	propfind_context_init (&pfctx);
 	
  propfind_start:	
@@ -1717,19 +1848,20 @@ http_get_file_info (HttpContext *context, GnomeVFSFileInfo *info, char **etag)
 	DEBUG_HTTP ("%d, %d", res, ne_get_status (req)->code);
 	ne_propfind_destroy (pfh);
 	
-	/* some server (eg. gws) colse the connection on an unknown command 
-	   so fall back to head for these too */
-	if (result != GNOME_VFS_ERROR_GENERIC  && 
-	    result != GNOME_VFS_ERROR_NOT_SUPPORTED) {
+	/* Let's be very cautious here! If the server doesn't respond with a 
+	   207 here just fall back to HEAD because some server server (eg. gws) 
+	   close the connection on an unknown command, most (stupid) php scripts
+	   will treat PROPFIND as GET and some servers may deny us PROFIND but
+	   allow us HEAD. */
+	if (res == NE_OK && ne_get_status (req)->code == 207) {
 		
-		if (result == GNOME_VFS_OK) {
-			gnome_vfs_file_info_copy (info, pfctx.target);
-			
-			if (etag && pfctx.etag)
-				*etag = pfctx.etag;
-			pfctx.etag = NULL;
+
+		if (pfctx.target != NULL) {
+			gnome_vfs_file_info_copy (info, pfctx.target);	
+		} else {
+			result = GNOME_VFS_ERROR_NOT_FOUND;
 		}
-			
+		
 		propfind_context_clear (&pfctx);
 		return result;
 	}
@@ -1739,35 +1871,19 @@ http_get_file_info (HttpContext *context, GnomeVFSFileInfo *info, char **etag)
 	
  head_start:	
 	req  = ne_request_create (context->session, "HEAD", context->path);
-			
-	ne_add_response_header_handler (req, "Last-Modified", 
-					(ne_header_handler) set_last_modified,
-		   			info);
-	
-	ne_add_response_header_handler (req, "Content-Length", 
-					(ne_header_handler) set_content_length,
-					info);
-	
-	ne_add_response_header_handler (req, "Content-Type",
-					(ne_header_handler) set_content_type,
-					info);
-	
-	ne_add_response_header_handler (req, "Date",
-					(ne_header_handler) set_access_time,
-					info);
 
-	ne_add_response_header_handler (req, "ETag",
-					(ne_header_handler) set_etag,
-					&_etag);
-			
+	add_default_header_handlers (req, info);
 		
 	res = ne_request_dispatch (req);
 
 	if (res == NE_REDIRECT) {
 		result = http_follow_redirect (context);
 		
-		if (result == GNOME_VFS_OK)
+		if (result == GNOME_VFS_OK) {
+			ne_request_destroy (req);
+			req = NULL;
 			goto head_start;
+		}
 	}
 
 
@@ -1784,19 +1900,28 @@ http_get_file_info (HttpContext *context, GnomeVFSFileInfo *info, char **etag)
 		info->flags = GNOME_VFS_FILE_FLAGS_NONE;
 		
 		info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
-		
-		if (etag && _etag) {
-			*etag = _etag;
-			_etag = NULL;
+
+		/* work-around for broken icecast server */
+		if (! g_ascii_strcasecmp (info->mime_type, "audio/mpeg")) {
+			ne_close_connection (ne_get_session (req));
 		}
+		
 	}
 
-	if (_etag)
-		g_free (_etag);
-	
 	return result;
 }
 
+static void assure_trailing_slash (HttpContext *context)
+{
+	char *tofree;
+
+	if (! ne_path_has_trailing_slash (context->path)) {
+		tofree = context->path;
+		context->path = g_strconcat (tofree, "/", NULL);
+		g_free (tofree);
+	}
+}
+
 static GnomeVFSResult
 http_list_directory (HttpContext *context, PropfindContext *pfctx)
 {
@@ -1804,31 +1929,18 @@ http_list_directory (HttpContext *context, PropfindContext *pfctx)
 	ne_propfind_handler *pfh;
 	ne_request *req;
 	int res;
-	char *tofree;
 
 	DEBUG_HTTP_CONTEXT (context);
 	
-	tofree = NULL;
 	propfind_context_init (pfctx);
+	pfctx->path = context->path;
 	
  propfind_start:	
-	if (ne_path_has_trailing_slash (context->path)) {
-		pfctx->path = context->path;
-	} else {
-		tofree = g_strconcat (context->path, "/", NULL);
-		pfctx->path = tofree;
-	}
-	
 	pfctx->include_target = TRUE;
-
+	
 	pfh = ne_propfind_create (context->session, context->path, 1);
 	res = ne_propfind_named (pfh, file_info_props, propfind_result, pfctx);
 	
-	if (tofree != NULL) {
-		g_free (tofree);
-		tofree = NULL;
-	}
-
 	if (res == NE_REDIRECT) {
 		
 		ne_propfind_destroy (pfh);
@@ -1846,22 +1958,26 @@ http_list_directory (HttpContext *context, PropfindContext *pfctx)
 	result 	= resolve_result (res, req);
 	
 	ne_propfind_destroy (pfh);
+
+	if (result == GNOME_VFS_OK && pfctx->target == NULL) {
+		return GNOME_VFS_ERROR_NOT_FOUND;
+	}
 	
 	return result;
 }
 
 
 static GnomeVFSResult
-http_options (HttpContext *hctx, gint *dav_class)
+http_options (HttpContext *hctx)
 {
 	GnomeVFSResult  result;
 	ne_request     *req;
-	gint 		klass;
+	DavClass 	klass;
 	HttpMethods	methods;
 	int		res;
 	
  options_start:	
-	klass = -1;
+	klass = DAV_CLASS_NOT_SET;
 	methods = 0;
 	req = ne_request_create (hctx->session, "OPTIONS", hctx->path);
 	ne_add_response_header_handler (req, "DAV", set_dav_class, &klass);
@@ -1876,11 +1992,10 @@ http_options (HttpContext *hctx, gint *dav_class)
 	}
 
 	result = resolve_result (res, req);
+	ne_request_destroy (req);
 	
 	if (result == GNOME_VFS_OK) {
-		if (dav_class)
-			*dav_class = klass;
-	
+		
 		hctx->dav_class = klass;
 		hctx->methods = methods;
 	}
@@ -1902,7 +2017,7 @@ end_response (void *userdata, void *response,
 }
 
 static int
-dav_request (ne_request *req)
+dav_request (ne_request *req, gboolean allow_redirect)
 {
 	ne_207_parser   *p207;
 	ne_xml_parser   *p;
@@ -1919,7 +2034,7 @@ dav_request (ne_request *req)
 	ne_207_set_propstat_handlers (p207, NULL, NULL /* end_propstat */);
     
 	ne_add_response_body_reader (req, ne_accept_207, ne_xml_parse_v, p);
-    
+	
 	res = ne_request_dispatch (req);
 	status = (ne_status *) ne_get_status (req);
     
@@ -1930,7 +2045,8 @@ dav_request (ne_request *req)
 			status->code = error;
 			status->klass = error % 100;
 		}
-	} else if (status->klass != 2) {
+	} else if (status->klass != 2 &&
+		   (!allow_redirect || res != NE_REDIRECT)) {
 		return NE_ERROR;
 	}
 	
@@ -1945,11 +2061,11 @@ dav_request (ne_request *req)
 
 typedef struct {
 	
-	HttpContext      *context;
-	GnomeVFSOpenMode  mode;
-	GnomeVFSFileInfo  *info;
-	GnomeVFSFileOffset offset;
-	char *etag;
+	HttpContext        *context;
+	GnomeVFSOpenMode    mode;
+	GnomeVFSFileInfo   *info;
+	GnomeVFSFileOffset  offset;
+	char               *etag;
 	
 	/* Whether we do/can ranged gets or not */
 	gboolean can_range;
@@ -2050,7 +2166,7 @@ http_transfer_start_write (HttpFileHandle *handle)
 	
 	/* Check to see if we have any file information. If not fetch it!*/
 	if (handle->info->valid_fields == 0) {
-		result = http_get_file_info (hctx, handle->info, NULL);
+		result = http_get_file_info (hctx, handle->info);
 
 		if (result != GNOME_VFS_OK) {
 			return result;
@@ -2108,27 +2224,8 @@ get_start:
 						&range);
 	}
 	
-		
-	ne_add_response_header_handler (req, "Last-Modified", 
-					(ne_header_handler) set_last_modified,
-		   			handle->info);
-	
-	ne_add_response_header_handler (req, "Content-Length", 
-					(ne_header_handler) set_content_length,
-					handle->info);
-	
-	ne_add_response_header_handler (req, "Content-Type",
-					(ne_header_handler) set_content_type,
-					handle->info);
-	
-	ne_add_response_header_handler (req, "Date",
-					(ne_header_handler) set_access_time,
-					handle->info);
-	
-	ne_add_response_header_handler (req, "ETag",
-					(ne_header_handler) set_etag,
-					&(handle->etag));
-	
+
+	add_default_header_handlers (req, handle->info);
 	
 	res = ne_begin_request (req);
 	result = resolve_result (res, req);
@@ -2140,6 +2237,7 @@ get_start:
 	if (res != NE_OK && res != NE_REDIRECT) {
 		handle->transfer_state = TRANSFER_ERROR;
 		handle->last_error = result;
+		ne_request_destroy (req);
 		return result;
 	}
 	
@@ -2178,6 +2276,13 @@ get_start:
 			DEBUG_HTTP ("[GET] {ranged} disabled");
 			handle->can_range = FALSE;
 		}
+	
+		/* If we are in a GET we invoke the callback of received headers 
+		   right before reading the data because we might be in a 
+		   stream and we wanna have the headers callbac invoked ASAP
+		*/
+		neon_return_headers (req, NULL, status);
+
 		
 		handle->transfer_state = TRANSFER_READ;
 		handle->transfer.read = req;
@@ -2211,6 +2316,8 @@ do_open (GnomeVFSMethod 	*method,
 	HttpFileHandle *handle;
 	GnomeVFSResult result;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	_GNOME_VFS_METHOD_PARAM_CHECK (uri != NULL);
 	
@@ -2231,7 +2338,7 @@ do_open (GnomeVFSMethod 	*method,
 	hctx = handle->context;
 
 	if (mode & GNOME_VFS_OPEN_WRITE) {
-		if ((result = http_options (hctx, NULL)) != GNOME_VFS_OK) {
+		if ((result = http_options (hctx)) != GNOME_VFS_OK) {
 			http_file_handle_destroy (handle);
 			return result;
 		}
@@ -2271,6 +2378,7 @@ do_open (GnomeVFSMethod 	*method,
 	
 	*method_handle = (GnomeVFSMethodHandle *) handle;
 	
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2289,6 +2397,8 @@ do_create (GnomeVFSMethod	 *method,
 	int res;
 	ne_request *req;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 
 	if ((mode & GNOME_VFS_OPEN_READ))
@@ -2311,10 +2421,11 @@ do_create (GnomeVFSMethod	 *method,
 #endif
 	
 	if (exclusive == TRUE) {
-		result = http_get_file_info (hctx, handle->info, NULL);
+		result = http_get_file_info (hctx, handle->info);
 		
 		if (result != GNOME_VFS_ERROR_NOT_FOUND) {
 			http_file_handle_destroy (handle);
+			ne_request_destroy (req);
 			return GNOME_VFS_ERROR_FILE_EXISTS;
 		}
 	}
@@ -2323,6 +2434,7 @@ do_create (GnomeVFSMethod	 *method,
 	
 	res = ne_request_dispatch (req);
 	result 	= resolve_result (res, req);
+	ne_request_destroy (req);
 	
 	if (result == GNOME_VFS_OK && mode != GNOME_VFS_OPEN_NONE) {
 
@@ -2341,6 +2453,7 @@ do_create (GnomeVFSMethod	 *method,
 
 	*method_handle = (GnomeVFSMethodHandle *) handle;	
 
+	DEBUG_HTTP_FUNC (0);
 	return result;
 	
 }
@@ -2354,6 +2467,8 @@ do_close (GnomeVFSMethod 	*method,
 	HttpFileHandle *handle;
 	HttpContext    *ctx;
 	int res;
+
+	DEBUG_HTTP_FUNC (1);
 	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	
@@ -2381,10 +2496,12 @@ do_close (GnomeVFSMethod 	*method,
 		DEBUG_HTTP ("[PUT] returned: %d, %d (%s)", res, 
 				ne_get_status (req)->code, 
 				gnome_vfs_result_to_string (result));
+		ne_request_destroy (req);
 	}
 	
 	
 	http_file_handle_destroy (handle);
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2526,6 +2643,8 @@ do_seek (GnomeVFSMethod	      *method,
 	HttpFileHandle 		*handle;
 	GnomeVFSFileOffset  	 new_position;
 
+	DEBUG_HTTP_FUNC (1);
+	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	
 	handle = (HttpFileHandle *) method_handle;
@@ -2574,13 +2693,10 @@ do_seek (GnomeVFSMethod	      *method,
 	handle->offset = new_position;
 	
 	if (handle->transfer_state == TRANSFER_READ) {
-				
 		http_transfer_abort (handle);
-		result = http_transfer_start (handle);
 	}
 	
-	
-	
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2629,15 +2745,22 @@ do_open_directory (GnomeVFSMethod 	    *method,
 	GnomeVFSResult  result;
 	HttpContext    *hctx;
 	PropfindContext *pfctx;
+
+	DEBUG_HTTP_FUNC (1);
+
+	if (scheme_is_dav (uri) == FALSE) {
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
 	
 	result = http_context_open  (uri, &hctx);
 	
 	if (result != GNOME_VFS_OK)
 		return result;
-	
-	result = http_options (hctx, NULL);
 
-	if (result != GNOME_VFS_OK || hctx->dav_class == -1) {
+	assure_trailing_slash (hctx);
+	result = http_options (hctx);
+
+	if (result != GNOME_VFS_OK || hctx->dav_class == NO_DAV) {
 		http_context_free (hctx);
 		
 		if (result != GNOME_VFS_OK)
@@ -2666,6 +2789,7 @@ do_open_directory (GnomeVFSMethod 	    *method,
 		*method_handle 	= (GnomeVFSMethodHandle  *) pfctx;
 	}
 	
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2676,6 +2800,7 @@ do_close_directory (GnomeVFSMethod 	 *method,
 {
 	PropfindContext *pfctx;
 	
+	DEBUG_HTTP_FUNC (1);
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 
 	pfctx = (PropfindContext *) method_handle;
@@ -2683,7 +2808,8 @@ do_close_directory (GnomeVFSMethod 	 *method,
 	propfind_context_clear (pfctx);
 	g_free (pfctx);
 	
-	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	DEBUG_HTTP_FUNC (0);
+	return GNOME_VFS_OK;
 }
 
 static GnomeVFSResult
@@ -2728,7 +2854,7 @@ do_get_file_info (GnomeVFSMethod 		*method,
 	if (result != GNOME_VFS_OK)
 		return result;
 		
-	result = http_get_file_info (hctx, file_info, NULL);
+	result = http_get_file_info (hctx, file_info);
 	
 	http_context_free (hctx);
 	
@@ -2747,6 +2873,8 @@ do_get_file_info_from_handle (GnomeVFSMethod 		*method,
 	HttpFileHandle *handle;
 	GnomeVFSResult  result;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	_GNOME_VFS_METHOD_PARAM_CHECK (method_handle != NULL);
 	
 	handle = (HttpFileHandle *) method_handle;
@@ -2756,7 +2884,7 @@ do_get_file_info_from_handle (GnomeVFSMethod 		*method,
 		return GNOME_VFS_OK;
 	}
 	
-	result = http_get_file_info (handle->context, handle->info, NULL);
+	result = http_get_file_info (handle->context, handle->info);
 
 	if (result == GNOME_VFS_OK) {
 		gnome_vfs_file_info_copy (file_info, handle->info);
@@ -2788,6 +2916,12 @@ do_make_directory (GnomeVFSMethod  *method,
 	ne_request       *req;
 	int 		  res;
 	
+	DEBUG_HTTP_FUNC (1);
+
+	if (scheme_is_dav (uri) == FALSE) {
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
+		
 	uri_parent = gnome_vfs_uri_get_parent (uri);
 	
 	result = http_context_open (uri_parent, &hctx);
@@ -2795,9 +2929,9 @@ do_make_directory (GnomeVFSMethod  *method,
 	if (result != GNOME_VFS_OK)
 		return result;
 	
-	result = http_options (hctx, NULL);
+	result = http_options (hctx);
 	
-	if (result != GNOME_VFS_OK || (hctx->dav_class == -1)) {
+	if (result != GNOME_VFS_OK || (hctx->dav_class == NO_DAV)) {
 
 		if (result == GNOME_VFS_OK)
 			result = GNOME_VFS_ERROR_NOT_SUPPORTED;
@@ -2815,6 +2949,8 @@ do_make_directory (GnomeVFSMethod  *method,
 		result = http_follow_redirect (hctx);
 		
 		if (result == GNOME_VFS_OK)
+			ne_request_destroy (req);
+			req = NULL;
 			goto mkcol_start;
 	} else if (res == NE_OK) {
 		const ne_status *status = ne_get_status (req);
@@ -2831,6 +2967,7 @@ do_make_directory (GnomeVFSMethod  *method,
  out:	
 	gnome_vfs_uri_unref (uri_parent);
 	http_context_free (hctx);
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2845,10 +2982,18 @@ do_remove_directory (GnomeVFSMethod  *method,
 	ne_request     *req;
 	int  		res;
 	
+	DEBUG_HTTP_FUNC (1);
+
+	if (scheme_is_dav (uri) == FALSE) {
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
+	
 	result = http_context_open  (uri, &hctx);
 	
 	if (result != GNOME_VFS_OK)
 		return result;
+
+	assure_trailing_slash (hctx);
 	
 	propfind_context_init (&pfctx);
 	result = http_list_directory (hctx, &pfctx);
@@ -2867,7 +3012,7 @@ do_remove_directory (GnomeVFSMethod  *method,
 	}
 	
 	req = ne_request_create (hctx->session, "DELETE", hctx->path);
-	res = dav_request (req);
+	res = dav_request (req, FALSE);
 	
 	result = resolve_result (res, req);
 	ne_request_destroy (req);
@@ -2875,9 +3020,21 @@ do_remove_directory (GnomeVFSMethod  *method,
  out:	
 	propfind_context_clear (&pfctx);
 	http_context_free (hctx);
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
+static GnomeVFSURI *
+resolve_schema_alias (GnomeVFSURI *uri)
+{
+	GnomeVFSURI *resolved;
+
+	resolved = gnome_vfs_uri_dup (uri);
+	g_free (resolved->method_string);
+	resolved->method_string = g_strdup (resolve_alias (gnome_vfs_uri_get_scheme (uri)));
+	return resolved;
+}
+
 
 static GnomeVFSResult
 do_move (GnomeVFSMethod  *method,
@@ -2891,7 +3048,13 @@ do_move (GnomeVFSMethod  *method,
 	char	    *dest;
 	ne_request  *req;
 	int 	     res;
+	GnomeVFSURI *dest_uri;
+	
+	DEBUG_HTTP_FUNC (1);
 
+	if (scheme_is_dav (old_uri) == FALSE) {
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
 	
 	if (! http_session_uri_equal (old_uri, new_uri)) {
 		return GNOME_VFS_ERROR_NOT_SAME_FILE_SYSTEM;
@@ -2901,22 +3064,37 @@ do_move (GnomeVFSMethod  *method,
 	
 	if (result != GNOME_VFS_OK)
 		return result;
-	
-	dest = gnome_vfs_uri_to_string (new_uri, GNOME_VFS_URI_HIDE_USER_NAME | 
+
+	dest_uri = resolve_schema_alias (new_uri);
+	dest = gnome_vfs_uri_to_string (dest_uri, GNOME_VFS_URI_HIDE_USER_NAME | 
 					GNOME_VFS_URI_HIDE_PASSWORD);
+	gnome_vfs_uri_unref (dest_uri);
 	
 
+ head_start:
 	req = ne_request_create (hctx->session, "MOVE", hctx->path);
 	
 	ne_add_request_header (req, "Destination", dest);
     	ne_add_request_header (req, "Overwrite", force_replace ? "T" : "F");
-
 	
-	res = dav_request (req);
-	result = resolve_result (res, req);
+	
+	res = dav_request (req, TRUE);
+	if (res == NE_REDIRECT) {
+		result = http_follow_redirect (hctx);
 
+		if (result == GNOME_VFS_OK) {
+			ne_request_destroy (req);
+			req = NULL;
+			goto head_start;
+		}
+	} else {
+		result = resolve_result (res, req);
+	}
+
+	ne_request_destroy (req);
 	http_context_free (hctx);
 	
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2931,13 +3109,14 @@ do_unlink (GnomeVFSMethod  *method,
 	ne_request       *req;
 	int 		  res;
 	
+	DEBUG_HTTP_FUNC (1);
 	result = http_context_open (uri, &hctx);
 	
 	if (result != GNOME_VFS_OK)
 		return result;
 	
 	file_info = gnome_vfs_file_info_new ();
-	result = http_get_file_info (hctx, file_info, NULL);
+	result = http_get_file_info (hctx, file_info);
 	
 	if (result != GNOME_VFS_OK) {
 		goto out;
@@ -2949,7 +3128,7 @@ do_unlink (GnomeVFSMethod  *method,
 	}
 	
 	req = ne_request_create (hctx->session, "DELETE", hctx->path);
-	res = dav_request (req);
+	res = dav_request (req, FALSE);
 
 	result = resolve_result (res, req);
 	ne_request_destroy (req);
@@ -2957,6 +3136,7 @@ do_unlink (GnomeVFSMethod  *method,
  out:	
 	http_context_free (hctx);
 	gnome_vfs_file_info_unref (file_info);
+	DEBUG_HTTP_FUNC (0);
 	return result;
 }
 
@@ -2983,6 +3163,8 @@ do_set_file_info (GnomeVFSMethod 		*method,
 	GnomeVFSURI *parent_uri, *new_uri;
 	GnomeVFSResult result;
 	
+	DEBUG_HTTP_FUNC (1);
+	
 	if ((mask & ~(GNOME_VFS_SET_FILE_INFO_NAME)) != 0) {
 		return GNOME_VFS_ERROR_NOT_SUPPORTED;
 	}
@@ -3004,6 +3186,8 @@ do_set_file_info (GnomeVFSMethod 		*method,
 	
 	gnome_vfs_uri_unref (new_uri);
 	
+	DEBUG_HTTP_FUNC (0);
+	
 	return result;
 }
 
@@ -3101,6 +3285,7 @@ GnomeVFSMethod *
 vfs_module_init (const char *method_name, const char *args)
 {
 	proxy_init ();
+	/* ne_debug_init (stdout, 0xfffe); */
 	neon_session_pool_init ();
 	http_auth_cache_init ();
 	quick_allow_lookup_init ();
